@@ -1,16 +1,17 @@
 import atexit
 from distutils.version import StrictVersion
-import glob
-import uuid
+
 import numpy as np
 import os
 import subprocess
-from typing import Dict, List, Optional, Any, Tuple
+from typing import Dict, List, Optional, Tuple, Mapping as MappingType
 
 import mlagents_envs
 
 from mlagents_envs.logging_util import get_logger
-from mlagents_envs.side_channel.side_channel import SideChannel, IncomingMessage
+from mlagents_envs.side_channel.side_channel import SideChannel
+from mlagents_envs.side_channel.side_channel_manager import SideChannelManager
+from mlagents_envs import env_utils
 
 from mlagents_envs.base_env import (
     BaseEnv,
@@ -19,6 +20,7 @@ from mlagents_envs.base_env import (
     BehaviorSpec,
     BehaviorName,
     AgentId,
+    BehaviorMapping,
 )
 from mlagents_envs.timers import timed, hierarchical_timer
 from mlagents_envs.exception import (
@@ -43,17 +45,12 @@ from mlagents_envs.communicator_objects.unity_rl_initialization_input_pb2 import
 from mlagents_envs.communicator_objects.unity_input_pb2 import UnityInputProto
 
 from .rpc_communicator import RpcCommunicator
-from sys import platform
 import signal
-import struct
 
 logger = get_logger(__name__)
 
 
 class UnityEnvironment(BaseEnv):
-    SCALAR_ACTION_TYPES = (int, np.int32, np.int64, float, np.float32, np.float64)
-    SINGLE_BRAIN_ACTION_TYPES = SCALAR_ACTION_TYPES + (list, np.ndarray)
-
     # Communication protocol version.
     # When connecting to C#, this must be compatible with Academy.k_ApiVersion.
     # We follow semantic versioning on the communication version, so existing
@@ -70,7 +67,7 @@ class UnityEnvironment(BaseEnv):
     BASE_ENVIRONMENT_PORT = 5005
 
     # Command line argument used to pass the port to the executable environment.
-    PORT_COMMAND_LINE_ARG = "--mlagents-port"
+    _PORT_COMMAND_LINE_ARG = "--mlagents-port"
 
     @staticmethod
     def _raise_version_exception(unity_com_ver: str) -> None:
@@ -82,7 +79,7 @@ class UnityEnvironment(BaseEnv):
         )
 
     @staticmethod
-    def check_communication_compatibility(
+    def _check_communication_compatibility(
         unity_com_ver: str, python_api_version: str, unity_package_version: str
     ) -> bool:
         unity_communicator_version = StrictVersion(unity_com_ver)
@@ -114,13 +111,13 @@ class UnityEnvironment(BaseEnv):
         return True
 
     @staticmethod
-    def get_capabilities_proto() -> UnityRLCapabilitiesProto:
+    def _get_capabilities_proto() -> UnityRLCapabilitiesProto:
         capabilities = UnityRLCapabilitiesProto()
         capabilities.baseRLCapabilities = True
         return capabilities
 
     @staticmethod
-    def warn_csharp_base_capabitlities(
+    def _warn_csharp_base_capabilities(
         caps: UnityRLCapabilitiesProto, unity_package_ver: str, python_package_ver: str
     ) -> None:
         if not caps.baseRLCapabilities:
@@ -141,8 +138,9 @@ class UnityEnvironment(BaseEnv):
         seed: int = 0,
         no_graphics: bool = False,
         timeout_wait: int = 60,
-        args: Optional[List[str]] = None,
+        additional_args: Optional[List[str]] = None,
         side_channels: Optional[List[SideChannel]] = None,
+        log_folder: Optional[str] = None,
     ):
         """
         Starts a new unity environment and establishes a connection with the environment.
@@ -157,34 +155,28 @@ class UnityEnvironment(BaseEnv):
         :int timeout_wait: Time (in seconds) to wait for connection from environment.
         :list args: Addition Unity command line arguments
         :list side_channels: Additional side channel for no-rl communication with Unity
+        :str log_folder: Optional folder to write the Unity Player log file into.  Requires absolute path.
         """
-        args = args or []
         atexit.register(self._close)
+        self._additional_args = additional_args or []
+        self._no_graphics = no_graphics
         # If base port is not specified, use BASE_ENVIRONMENT_PORT if we have
         # an environment, otherwise DEFAULT_EDITOR_PORT
         if base_port is None:
             base_port = (
                 self.BASE_ENVIRONMENT_PORT if file_name else self.DEFAULT_EDITOR_PORT
             )
-        self.port = base_port + worker_id
+        self._port = base_port + worker_id
         self._buffer_size = 12000
         # If true, this means the environment was successfully loaded
         self._loaded = False
         # The process that is started. If None, no process was started
-        self.proc1 = None
-        self.timeout_wait: int = timeout_wait
-        self.communicator = self.get_communicator(worker_id, base_port, timeout_wait)
-        self.worker_id = worker_id
-        self.side_channels: Dict[uuid.UUID, SideChannel] = {}
-        if side_channels is not None:
-            for _sc in side_channels:
-                if _sc.channel_id in self.side_channels:
-                    raise UnityEnvironmentException(
-                        "There cannot be two side channels with the same channel id {0}.".format(
-                            _sc.channel_id
-                        )
-                    )
-                self.side_channels[_sc.channel_id] = _sc
+        self._proc1 = None
+        self._timeout_wait: int = timeout_wait
+        self._communicator = self._get_communicator(worker_id, base_port, timeout_wait)
+        self._worker_id = worker_id
+        self._side_channel_manager = SideChannelManager(side_channels)
+        self._log_folder = log_folder
 
         # If the environment name is None, a new environment will not be launched
         # and the communicator will directly try to connect to an existing unity environment.
@@ -195,10 +187,16 @@ class UnityEnvironment(BaseEnv):
                 "the worker-id must be 0 in order to connect with the Editor."
             )
         if file_name is not None:
-            self.executable_launcher(file_name, no_graphics, args)
+            try:
+                self._proc1 = env_utils.launch_executable(
+                    file_name, self._executable_args()
+                )
+            except UnityEnvironmentException:
+                self._close(0)
+                raise
         else:
             logger.info(
-                f"Listening on port {self.port}. "
+                f"Listening on port {self._port}. "
                 f"Start training by pressing the Play button in the Unity Editor."
             )
         self._loaded = True
@@ -207,16 +205,16 @@ class UnityEnvironment(BaseEnv):
             seed=seed,
             communication_version=self.API_VERSION,
             package_version=mlagents_envs.__version__,
-            capabilities=UnityEnvironment.get_capabilities_proto(),
+            capabilities=UnityEnvironment._get_capabilities_proto(),
         )
         try:
-            aca_output = self.send_academy_parameters(rl_init_parameters_in)
+            aca_output = self._send_academy_parameters(rl_init_parameters_in)
             aca_params = aca_output.rl_initialization_output
         except UnityTimeOutException:
             self._close(0)
             raise
 
-        if not UnityEnvironment.check_communication_compatibility(
+        if not UnityEnvironment._check_communication_compatibility(
             aca_params.communication_version,
             UnityEnvironment.API_VERSION,
             aca_params.package_version,
@@ -224,7 +222,7 @@ class UnityEnvironment(BaseEnv):
             self._close(0)
             UnityEnvironment._raise_version_exception(aca_params.communication_version)
 
-        UnityEnvironment.warn_csharp_base_capabitlities(
+        UnityEnvironment._warn_csharp_base_capabilities(
             aca_params.capabilities,
             aca_params.package_version,
             UnityEnvironment.API_VERSION,
@@ -237,97 +235,22 @@ class UnityEnvironment(BaseEnv):
         self._update_behavior_specs(aca_output)
 
     @staticmethod
-    def get_communicator(worker_id, base_port, timeout_wait):
+    def _get_communicator(worker_id, base_port, timeout_wait):
         return RpcCommunicator(worker_id, base_port, timeout_wait)
 
-    @staticmethod
-    def validate_environment_path(env_path: str) -> Optional[str]:
-        # Strip out executable extensions if passed
-        env_path = (
-            env_path.strip()
-            .replace(".app", "")
-            .replace(".exe", "")
-            .replace(".x86_64", "")
-            .replace(".x86", "")
-        )
-        true_filename = os.path.basename(os.path.normpath(env_path))
-        logger.debug("The true file name is {}".format(true_filename))
-
-        if not (glob.glob(env_path) or glob.glob(env_path + ".*")):
-            return None
-
-        cwd = os.getcwd()
-        launch_string = None
-        true_filename = os.path.basename(os.path.normpath(env_path))
-        if platform == "linux" or platform == "linux2":
-            candidates = glob.glob(os.path.join(cwd, env_path) + ".x86_64")
-            if len(candidates) == 0:
-                candidates = glob.glob(os.path.join(cwd, env_path) + ".x86")
-            if len(candidates) == 0:
-                candidates = glob.glob(env_path + ".x86_64")
-            if len(candidates) == 0:
-                candidates = glob.glob(env_path + ".x86")
-            if len(candidates) > 0:
-                launch_string = candidates[0]
-
-        elif platform == "darwin":
-            candidates = glob.glob(
-                os.path.join(cwd, env_path + ".app", "Contents", "MacOS", true_filename)
+    def _executable_args(self) -> List[str]:
+        args: List[str] = []
+        if self._no_graphics:
+            args += ["-nographics", "-batchmode"]
+        args += [UnityEnvironment._PORT_COMMAND_LINE_ARG, str(self._port)]
+        if self._log_folder:
+            log_file_path = os.path.join(
+                self._log_folder, f"Player-{self._worker_id}.log"
             )
-            if len(candidates) == 0:
-                candidates = glob.glob(
-                    os.path.join(env_path + ".app", "Contents", "MacOS", true_filename)
-                )
-            if len(candidates) == 0:
-                candidates = glob.glob(
-                    os.path.join(cwd, env_path + ".app", "Contents", "MacOS", "*")
-                )
-            if len(candidates) == 0:
-                candidates = glob.glob(
-                    os.path.join(env_path + ".app", "Contents", "MacOS", "*")
-                )
-            if len(candidates) > 0:
-                launch_string = candidates[0]
-        elif platform == "win32":
-            candidates = glob.glob(os.path.join(cwd, env_path + ".exe"))
-            if len(candidates) == 0:
-                candidates = glob.glob(env_path + ".exe")
-            if len(candidates) > 0:
-                launch_string = candidates[0]
-        return launch_string
-
-    def executable_launcher(self, file_name, no_graphics, args):
-        launch_string = self.validate_environment_path(file_name)
-        if launch_string is None:
-            self._close(0)
-            raise UnityEnvironmentException(
-                f"Couldn't launch the {file_name} environment. Provided filename does not match any environments."
-            )
-        else:
-            logger.debug("This is the launch string {}".format(launch_string))
-            # Launch Unity environment
-            subprocess_args = [launch_string]
-            if no_graphics:
-                subprocess_args += ["-nographics", "-batchmode"]
-            subprocess_args += [UnityEnvironment.PORT_COMMAND_LINE_ARG, str(self.port)]
-            subprocess_args += args
-            try:
-                self.proc1 = subprocess.Popen(
-                    subprocess_args,
-                    # start_new_session=True means that signals to the parent python process
-                    # (e.g. SIGINT from keyboard interrupt) will not be sent to the new process on POSIX platforms.
-                    # This is generally good since we want the environment to have a chance to shutdown,
-                    # but may be undesirable in come cases; if so, we'll add a command-line toggle.
-                    # Note that on Windows, the CTRL_C signal will still be sent.
-                    start_new_session=True,
-                )
-            except PermissionError as perm:
-                # This is likely due to missing read or execute permissions on file.
-                raise UnityEnvironmentException(
-                    f"Error when trying to launch environment - make sure "
-                    f"permissions are set correctly. For example "
-                    f'"chmod -R 755 {launch_string}"'
-                ) from perm
+            args += ["-logFile", log_file_path]
+        # Add in arguments passed explicitly by the user.
+        args += self._additional_args
+        return args
 
     def _update_behavior_specs(self, output: UnityOutputProto) -> None:
         init_output = output.rl_initialization_output
@@ -356,11 +279,11 @@ class UnityEnvironment(BaseEnv):
                     DecisionSteps.empty(self._env_specs[brain_name]),
                     TerminalSteps.empty(self._env_specs[brain_name]),
                 )
-        self._parse_side_channel_message(self.side_channels, output.side_channel)
+        self._side_channel_manager.process_side_channel_message(output.side_channel)
 
     def reset(self) -> None:
         if self._loaded:
-            outputs = self.communicator.exchange(self._generate_reset_input())
+            outputs = self._communicator.exchange(self._generate_reset_input())
             if outputs is None:
                 raise UnityCommunicatorStoppedException("Communicator has exited.")
             self._update_behavior_specs(outputs)
@@ -388,7 +311,7 @@ class UnityEnvironment(BaseEnv):
                 ].create_empty_action(n_agents)
         step_input = self._generate_step_input(self._env_actions)
         with hierarchical_timer("communicator.exchange"):
-            outputs = self.communicator.exchange(step_input)
+            outputs = self._communicator.exchange(step_input)
         if outputs is None:
             raise UnityCommunicatorStoppedException("Communicator has exited.")
         self._update_behavior_specs(outputs)
@@ -396,14 +319,15 @@ class UnityEnvironment(BaseEnv):
         self._update_state(rl_output)
         self._env_actions.clear()
 
-    def get_behavior_names(self):
-        return list(self._env_specs.keys())
+    @property
+    def behavior_specs(self) -> MappingType[str, BehaviorSpec]:
+        return BehaviorMapping(self._env_specs)
 
     def _assert_behavior_exists(self, behavior_name: str) -> None:
         if behavior_name not in self._env_specs:
             raise UnityActionException(
-                "The group {0} does not correspond to an existing agent group "
-                "in the environment".format(behavior_name)
+                f"The group {behavior_name} does not correspond to an existing "
+                f"agent group in the environment"
             )
 
     def set_actions(self, behavior_name: BehaviorName, action: np.ndarray) -> None:
@@ -415,9 +339,9 @@ class UnityEnvironment(BaseEnv):
         expected_shape = (len(self._env_state[behavior_name][0]), spec.action_size)
         if action.shape != expected_shape:
             raise UnityActionException(
-                "The behavior {0} needs an input of dimension {1} for "
-                "(<number of agents>, <action size>) but received input of "
-                "dimension {2}".format(behavior_name, expected_shape, action.shape)
+                f"The behavior {behavior_name} needs an input of dimension "
+                f"{expected_shape} for (<number of agents>, <action size>) but "
+                f"received input of dimension {action.shape}"
             )
         if action.dtype != expected_type:
             action = action.astype(expected_type)
@@ -433,10 +357,9 @@ class UnityEnvironment(BaseEnv):
         expected_shape = (spec.action_size,)
         if action.shape != expected_shape:
             raise UnityActionException(
-                f"The Agent {0} with BehaviorName {1} needs an input of dimension "
-                f"{2} but received input of dimension {3}".format(
-                    agent_id, behavior_name, expected_shape, action.shape
-                )
+                f"The Agent {agent_id} with BehaviorName {behavior_name} needs "
+                f"an input of dimension {expected_shape} but received input of "
+                f"dimension {action.shape}"
             )
         expected_type = np.float32 if spec.is_action_continuous() else np.int32
         if action.dtype != expected_type:
@@ -464,10 +387,6 @@ class UnityEnvironment(BaseEnv):
         self._assert_behavior_exists(behavior_name)
         return self._env_state[behavior_name]
 
-    def get_behavior_spec(self, behavior_name: BehaviorName) -> BehaviorSpec:
-        self._assert_behavior_exists(behavior_name)
-        return self._env_specs[behavior_name]
-
     def close(self):
         """
         Sends a shutdown signal to the unity environment, and closes the socket connection.
@@ -485,91 +404,22 @@ class UnityEnvironment(BaseEnv):
             force-killing it.  Defaults to `self.timeout_wait`.
         """
         if timeout is None:
-            timeout = self.timeout_wait
+            timeout = self._timeout_wait
         self._loaded = False
-        self.communicator.close()
-        if self.proc1 is not None:
+        self._communicator.close()
+        if self._proc1 is not None:
             # Wait a bit for the process to shutdown, but kill it if it takes too long
             try:
-                self.proc1.wait(timeout=timeout)
-                signal_name = self.returncode_to_signal_name(self.proc1.returncode)
+                self._proc1.wait(timeout=timeout)
+                signal_name = self._returncode_to_signal_name(self._proc1.returncode)
                 signal_name = f" ({signal_name})" if signal_name else ""
-                return_info = f"Environment shut down with return code {self.proc1.returncode}{signal_name}."
+                return_info = f"Environment shut down with return code {self._proc1.returncode}{signal_name}."
                 logger.info(return_info)
             except subprocess.TimeoutExpired:
                 logger.info("Environment timed out shutting down. Killing...")
-                self.proc1.kill()
+                self._proc1.kill()
             # Set to None so we don't try to close multiple times.
-            self.proc1 = None
-
-    @classmethod
-    def _flatten(cls, arr: Any) -> List[float]:
-        """
-        Converts arrays to list.
-        :param arr: numpy vector.
-        :return: flattened list.
-        """
-        if isinstance(arr, cls.SCALAR_ACTION_TYPES):
-            arr = [float(arr)]
-        if isinstance(arr, np.ndarray):
-            arr = arr.tolist()
-        if len(arr) == 0:
-            return arr
-        if isinstance(arr[0], np.ndarray):
-            # pylint: disable=no-member
-            arr = [item for sublist in arr for item in sublist.tolist()]
-        if isinstance(arr[0], list):
-            # pylint: disable=not-an-iterable
-            arr = [item for sublist in arr for item in sublist]
-        arr = [float(x) for x in arr]
-        return arr
-
-    @staticmethod
-    def _parse_side_channel_message(
-        side_channels: Dict[uuid.UUID, SideChannel], data: bytes
-    ) -> None:
-        offset = 0
-        while offset < len(data):
-            try:
-                channel_id = uuid.UUID(bytes_le=bytes(data[offset : offset + 16]))
-                offset += 16
-                message_len, = struct.unpack_from("<i", data, offset)
-                offset = offset + 4
-                message_data = data[offset : offset + message_len]
-                offset = offset + message_len
-            except Exception:
-                raise UnityEnvironmentException(
-                    "There was a problem reading a message in a SideChannel. "
-                    "Please make sure the version of MLAgents in Unity is "
-                    "compatible with the Python version."
-                )
-            if len(message_data) != message_len:
-                raise UnityEnvironmentException(
-                    "The message received by the side channel {0} was "
-                    "unexpectedly short. Make sure your Unity Environment "
-                    "sending side channel data properly.".format(channel_id)
-                )
-            if channel_id in side_channels:
-                incoming_message = IncomingMessage(message_data)
-                side_channels[channel_id].on_message_received(incoming_message)
-            else:
-                logger.warning(
-                    "Unknown side channel data received. Channel type "
-                    ": {0}.".format(channel_id)
-                )
-
-    @staticmethod
-    def _generate_side_channel_data(
-        side_channels: Dict[uuid.UUID, SideChannel]
-    ) -> bytearray:
-        result = bytearray()
-        for channel_id, channel in side_channels.items():
-            for message in channel.message_queue:
-                result += channel_id.bytes_le
-                result += struct.pack("<i", len(message))
-                result += message
-            channel.message_queue = []
-        return result
+            self._proc1 = None
 
     @timed
     def _generate_step_input(
@@ -584,30 +434,34 @@ class UnityEnvironment(BaseEnv):
                 action = AgentActionProto(vector_actions=vector_action[b][i])
                 rl_in.agent_actions[b].value.extend([action])
                 rl_in.command = STEP
-        rl_in.side_channel = bytes(self._generate_side_channel_data(self.side_channels))
-        return self.wrap_unity_input(rl_in)
+        rl_in.side_channel = bytes(
+            self._side_channel_manager.generate_side_channel_messages()
+        )
+        return self._wrap_unity_input(rl_in)
 
     def _generate_reset_input(self) -> UnityInputProto:
         rl_in = UnityRLInputProto()
         rl_in.command = RESET
-        rl_in.side_channel = bytes(self._generate_side_channel_data(self.side_channels))
-        return self.wrap_unity_input(rl_in)
+        rl_in.side_channel = bytes(
+            self._side_channel_manager.generate_side_channel_messages()
+        )
+        return self._wrap_unity_input(rl_in)
 
-    def send_academy_parameters(
+    def _send_academy_parameters(
         self, init_parameters: UnityRLInitializationInputProto
     ) -> UnityOutputProto:
         inputs = UnityInputProto()
         inputs.rl_initialization_input.CopyFrom(init_parameters)
-        return self.communicator.initialize(inputs)
+        return self._communicator.initialize(inputs)
 
     @staticmethod
-    def wrap_unity_input(rl_input: UnityRLInputProto) -> UnityInputProto:
+    def _wrap_unity_input(rl_input: UnityRLInputProto) -> UnityInputProto:
         result = UnityInputProto()
         result.rl_input.CopyFrom(rl_input)
         return result
 
     @staticmethod
-    def returncode_to_signal_name(returncode: int) -> Optional[str]:
+    def _returncode_to_signal_name(returncode: int) -> Optional[str]:
         """
         Try to convert return codes into their corresponding signal name.
         E.g. returncode_to_signal_name(-2) -> "SIGINT"
