@@ -22,6 +22,14 @@ class UnityCommunicationException(UnityException):
     pass
 
 
+class UnityCommunicatorStoppedException(UnityException):
+    """
+    Raised when communicator has stopped gracefully.
+    """
+
+    pass
+
+
 class UnityObservationException(UnityException):
     """
     Related to errors with receiving observations.
@@ -67,4 +75,4 @@ class UnityWorkerInUseException(UnityException):
 
     def __init__(self, worker_id):
         message = self.MESSAGE_TEMPLATE.format(str(worker_id))
-        super(UnityWorkerInUseException, self).__init__(message)
+        super().__init__(message)
